/******************************************************************************
 Web Notification Service (WNS): An open interface for a service
 by which a client may conduct asynchronous dialogues
 (message interchanges) with one or more other services.
 *******************************************************************************/
package org.n52.wns;

import java.io.File;
import java.io.FileInputStream;
import java.io.IOException;
import java.util.ArrayList;
import java.util.Iterator;
import java.util.Properties;
import java.util.logging.Level;
import java.util.logging.Logger;

import org.apache.xmlbeans.XmlOptions;
import org.n52.wns.WNSServiceException.ExceptionLevel;
import org.n52.wns.db.DAOFactory;
import org.x52North.wns.WNSConfigDocument;
import org.x52North.wns.WNSConfigDocument.WNSConfig;

/**
 * @author Johannes Echterhoff
 * @version 1.0
 * 
 */
public class WNSInitParamContainer {

	private static Logger logger = Logger.getLogger(WNSInitParamContainer.class
			.getName());

	private static WNSInitParamContainer instance = null;

	private WNSConfig wnsc = null;

	private Properties daoProps = null;

	private ExceptionLevel exceptionLevel = ExceptionLevel.PlainExceptions;

	private DAOFactory daoFactory = null;

	private String wnsSchemaLocation = null;

	private String ogcExcSchemaLocation = null;

	private String wnsUrl = null;

	private WNSInitParamContainer(WNSConfig wnsc, Properties daoProps) {

		this.wnsc = wnsc;
		this.daoProps = daoProps;

		if (wnsc.getServiceProperties().getExceptionLevel().equals(
				WNSConfig.ServiceProperties.ExceptionLevel.PLAIN_EXCEPTIONS)) {
			this.exceptionLevel = ExceptionLevel.PlainExceptions;
		} else {
			this.exceptionLevel = ExceptionLevel.DetailedExceptions;
		}

		this.wnsSchemaLocation = wnsc.getServiceProperties()
				.getWNSSchemaLocation();
		this.ogcExcSchemaLocation = wnsc.getServiceProperties()
				.getOWSSchemaLocation();
		this.wnsUrl = wnsc.getServiceProperties().getWNSURL();
	}

	/**
	 * @param wnsConfigLocation
	 * @return
	 * @throws WNSException
	 *             If the WNSInitParamContainer could not be initialised.
	 */
	public static synchronized WNSInitParamContainer createInstance(
			String wnsConfigLocation) throws WNSException {

		if (instance == null) {

			WNSConfigDocument wnscd = null;

			// parse WNSConfig
			try {
				wnscd = WNSConfigDocument.Factory.parse(new File(
						wnsConfigLocation));

				// validate WNSConfig
				ArrayList validationErrors = new ArrayList();
				XmlOptions validationOptions = new XmlOptions();
				validationOptions.setErrorListener(validationErrors);
				boolean isValid = wnscd.validate(validationOptions);

				// create XmlException with error-message if the XML is invalid.
				if (!isValid) {
					StringBuffer sb = new StringBuffer();
					Iterator iter = validationErrors.iterator();
					while (iter.hasNext()) {
						sb.append("[Validation-Error] " + iter.next() + "\n");
					}

					throw new WNSException("The WNSConfig is invalid!\n"
							+ sb.toString());

				} else {
					logger.info("The WNSConfig is valid.");
				}
			} catch (Exception e) {
				String message = "Failure while trying to read WNSConfig.";
				logger.log(Level.SEVERE, message, e);
				throw new WNSException(message, e);
			}

			// load properties for DAOFactory
			String daoFacPropsLocation = wnscd.getWNSConfig()
					.getServiceProperties().getDAOFactoryPropertiesLocation();
			Properties daoProps = new Properties();
			FileInputStream in;
			try {
				in = new FileInputStream(daoFacPropsLocation);
				daoProps.load(in);
				in.close();
			} catch (IOException e) {
				String message = "Could not load properties for DAOFactory.";
				logger.log(Level.SEVERE, message, e);
				throw new WNSException(message, e);
			}

			instance = new WNSInitParamContainer(wnscd.getWNSConfig(), daoProps);

			try {
				// initialise PMDAOFactory
				String driver = wnscd.getWNSConfig().getServiceProperties()
						.getDAOFactoryDriver();
				Class c = Class.forName(driver);
				instance.daoFactory = (DAOFactory) c.newInstance();
			} catch (Exception e) {
				String message = "Failed to initialise DAOFactory.";
				logger.log(Level.SEVERE, message, e);
				throw new WNSException(message, e);
			}
		}
		return instance;
	}

	public static WNSInitParamContainer getInstance() {
		return instance;
	}

	public ExceptionLevel getExceptionLevel() {
		return exceptionLevel;
	}

	public Properties getWNSDAOFactoryProperties() {
		return daoProps;
	}

	public String getOgcExceptionSchemaLocation() {
		return ogcExcSchemaLocation;
	}

	public String getWnsSchemaLocation() {
		return wnsSchemaLocation;
	}

	/**
	 * @return Returns the url of the WNS.
	 */
	public String getWnsUrl() {
		return wnsUrl;
	}

	/**
	 * @return Returns the DAOFactory.
	 */
	public DAOFactory getDAOFactory() {
		return daoFactory;
	}

	public WNSConfig getWnsConfig() {
		return wnsc;
	}

}
