/******************************************************************************
 Web Notification Service (WNS): An open interface for a service
 by which a client may conduct asynchronous dialogues
 (message interchanges) with one or more other services.
 *******************************************************************************/
package org.n52.wns;

/**
 * This class provides the necessary methods to handle messages
 * 
 * @author Dennis Dahlmann
 * @version 1.0
 */

import java.util.logging.Logger;

import net.opengis.wns.CommunicationMessageType;
import net.opengis.wns.DoCommunicationDocument;
import net.opengis.wns.DoCommunicationResponseDocument;
import net.opengis.wns.DoNotificationDocument;
import net.opengis.wns.DoNotificationResponseDocument;
import net.opengis.wns.GetMessageDocument;
import net.opengis.wns.GetMessageResponseDocument;
import net.opengis.wns.MessageStatusType;
import net.opengis.wns.NotificationMessageType;

import org.apache.xmlbeans.GDuration;
import org.n52.wns.communication.FaxHandler;
import org.n52.wns.communication.FaxHandlerFactory;
import org.n52.wns.communication.PhoneHandler;
import org.n52.wns.communication.PhoneHandlerFactory;
import org.n52.wns.mail.MailHandler;
import org.n52.wns.mail.MailHandlerFactory;
import org.n52.wns.sms.SMSHandler;
import org.n52.wns.sms.SMSHandlerFactory;
import org.n52.wns.xmpp.XMPPHandler;
import org.n52.wns.xmpp.XMPPHandlerJabber;
import org.x52North.wns.WNSConfigDocument.WNSConfig;
import org.xmldb.api.DatabaseManager;
import org.xmldb.api.base.Collection;
import org.xmldb.api.base.Database;
import org.xmldb.api.base.XMLDBException;
import org.xmldb.api.modules.CollectionManagementService;
import org.xmldb.api.modules.XMLResource;

public class WNSMessageHandler implements WNSMessageHandlerI {

	private static WNSMessageHandler instance = null;

	private Logger log = Logger.getLogger(WNSMessageHandler.class.getName());

	private MailHandler mailHandler = null;

	private SMSHandler smsHandler = null;

	private XMPPHandler xmppHandler = null;

	private PhoneHandler phoneHandler = null;

	private FaxHandler faxHandler = null;

	private WNSWebServiceHandler wsHandler = null;

	private WNSUserHandler userHandler = null;

	private WNSDatabaseThread thread = null;

	private GDuration maxTTLofMessage = null;

	private Collection col;

	private String connectionURL, dbURL, dbUser, dbPwd, dbName;

	private int messageID = 0;

	/**
	 * 
	 * @param userHandler
	 *            The handler to admin the user data
	 * @throws WNSServiceException
	 */
	public WNSMessageHandler(WNSUserHandler userHandler)
			throws WNSServiceException {
		WNSInitParamContainer initParams = WNSInitParamContainer.getInstance();
		WNSConfig wnsConfig = initParams.getWnsConfig();
		this.userHandler = userHandler;
		this.maxTTLofMessage = wnsConfig.getServiceProperties()
				.getMaxTTLOfMessages();
		dbURL = wnsConfig.getServiceProperties().getDataBaseProperties()
				.getURL();
		dbName = wnsConfig.getServiceProperties().getDataBaseProperties()
				.getDBName();
		dbUser = wnsConfig.getServiceProperties().getDataBaseProperties()
				.getUser();
		dbPwd = wnsConfig.getServiceProperties().getDataBaseProperties()
				.getPassword();

		connectionURL = "xmldb:exist://" + dbURL + "/xmlrpc" + "/db/" + dbName;

		String driver = "org.exist.xmldb.DatabaseImpl";
		Class cl;
		try {
			cl = Class.forName(driver);
			Database database = (Database) cl.newInstance();
			DatabaseManager.registerDatabase(database);
			col = DatabaseManager.getCollection(connectionURL, dbUser, dbPwd);
			if (col == null) {
				Collection root = DatabaseManager.getCollection(
						"xmldb:exist://" + dbURL + "/xmlrpc" + "/db", dbUser,
						dbPwd);
				CollectionManagementService mgtService = (CollectionManagementService) root
						.getService("CollectionManagementService", "1.0");
				col = mgtService.createCollection(dbName);
			}
			loadMessages();
		} catch (ClassNotFoundException e) {

		} catch (InstantiationException e) {

		} catch (IllegalAccessException e) {

		} catch (XMLDBException e) {

		} catch (WNSServiceException e) {

		}

		if (wnsConfig.getRegisteredHandlers().isSetEMailHandler()) {
			mailHandler = MailHandlerFactory.getInstance(wnsConfig
					.getRegisteredHandlers().getEMailHandler());
		}
		if (wnsConfig.getRegisteredHandlers().isSetSMSHandler()) {
			smsHandler = SMSHandlerFactory.getInstance(wnsConfig
					.getRegisteredHandlers().getSMSHandler(), wnsConfig
					.getServiceProperties().getWNSURL());
		}
		if (wnsConfig.getRegisteredHandlers().isSetPhoneHandler()) {
			phoneHandler = PhoneHandlerFactory.getInstance(wnsConfig
					.getRegisteredHandlers().getPhoneHandler(), wnsConfig
					.getServiceProperties().getWNSURL());
		}
		if (wnsConfig.getRegisteredHandlers().isSetFaxHandler()) {
			faxHandler = FaxHandlerFactory.getInstance(wnsConfig
					.getRegisteredHandlers().getFaxHandler(), wnsConfig
					.getServiceProperties().getWNSURL());
		}
		if (wnsConfig.getRegisteredHandlers().isSetXMPPHandler()) {
			try {
				xmppHandler = XMPPHandlerJabber.getInstance(wnsConfig
						.getRegisteredHandlers().getXMPPHandler());
			} catch (WNSServiceException e) {
				throw e;
			}
		}

		wsHandler = new WNSWebServiceHandler();

		thread = new WNSDatabaseThread(wnsConfig.getServiceProperties()
				.getDataBaseProperties().getCheckDBDuration(), col);

	}

	/**
	 * Singleton
	 * 
	 * @param userhandler
	 * @return
	 * @throws WNSServiceException
	 */
	public static synchronized WNSMessageHandler getInstance(
			WNSUserHandler userhandler) throws WNSServiceException {
		if (instance == null) {
			instance = new WNSMessageHandler(userhandler);
		}
		return instance;
	}

	/**
	 * Method to get the messageIDs from the database to increase from this
	 * number
	 * 
	 */
	private void loadMessages() throws WNSServiceException {
		try {
			String[] list = col.listResources();
			for (String message : list) {
				if (!message.equalsIgnoreCase("savedIDs")) {
					if (new Integer(message.substring(1)) > messageID) {
						messageID = new Integer(message.substring(1));
					}
				}
			}
		} catch (Exception e) {
			WNSServiceException s = new WNSServiceException(
					WNSServiceException.ExceptionLevel.PlainExceptions);
			s.addCodedException(
					WNSServiceException.ExceptionCode.DatabaseErrorOccured,
					null);
			throw s;

		}
		messageID++;
	}

	/**
	 * MAIN function to send a notification
	 */
	public DoNotificationResponseDocument doNotification(
			DoNotificationDocument dnd) throws WNSServiceException {
		DoNotificationResponseDocument response = DoNotificationResponseDocument.Factory
				.newInstance();
		org.x52North.wns.WNSUserDocument.WNSUser.SingleUser.User.Channel[] channel;
		XMLResource document;
		try {
			channel = userHandler.getMessageChannels(dnd.getDoNotification()
					.getUserID());
			boolean storeMessage = false;
			for (int i = 0; i < channel.length; i++) {
				switch (channel[i].getProtocol().intValue()) {
				case org.x52North.wns.WNSUserDocument.WNSUser.SingleUser.User.Channel.Protocol.INT_EMAIL:
					if (mailHandler != null) {
						mailHandler.sendNotificationMessage(dnd, channel[i]
								.getTargetArray());
					}
					break;
				case org.x52North.wns.WNSUserDocument.WNSUser.SingleUser.User.Channel.Protocol.INT_XMPP:
					if (xmppHandler != null) {

						xmppHandler.sendNotificationMessage(dnd, channel[i]
								.getTargetArray());
					}
					break;
				case org.x52North.wns.WNSUserDocument.WNSUser.SingleUser.User.Channel.Protocol.INT_SMS:
					if (smsHandler != null) {

						smsHandler.sendNotificationMessage(dnd, channel[i]
								.getTargetArray(), "N" + messageID);
						storeMessage = true;
					}
					break;
				case org.x52North.wns.WNSUserDocument.WNSUser.SingleUser.User.Channel.Protocol.INT_HTTP:
					if (wsHandler != null) {
						wsHandler.sendNotificationMessage(dnd, channel[i]
								.getTargetArray());
					}
					break;
				case org.x52North.wns.WNSUserDocument.WNSUser.SingleUser.User.Channel.Protocol.INT_PHONE:
					if (phoneHandler != null) {
						phoneHandler.sendNotificationMessage(dnd, channel[i]
								.getTargetArray(), "N" + messageID);
						storeMessage = true;
					}
					break;
				case org.x52North.wns.WNSUserDocument.WNSUser.SingleUser.User.Channel.Protocol.INT_FAX:
					if (faxHandler != null) {
						faxHandler.sendNotificationMessage(dnd, channel[i]
								.getTargetArray(), "N" + messageID);
						storeMessage = true;
					}
					break;

				default:
					response.addNewDoNotificationResponse().setStatus(
							MessageStatusType.MESSAGE_SENDING_FAILED);
					return response;
				}
			}
			if (storeMessage) {
				document = (XMLResource) col.createResource("N" + messageID,
						"XMLResource");
				document.setContent(dnd.getDoNotification().getMessage()
						.xmlText());
				col.storeResource(document);
				col.close();
				if (dnd.getDoNotification().getMaxTTLOfMessage()
						.compareToGDuration(maxTTLofMessage) < 0) {
					thread.add("N" + messageID, dnd.getDoNotification()
							.getMaxTTLOfMessage());
				} else {
					thread.add("N" + messageID, maxTTLofMessage);
				}
				messageID++;
			}
		} catch (WNSException e) {
			e.printStackTrace();
			response.addNewDoNotificationResponse().setStatus(
					MessageStatusType.MESSAGE_SENDING_FAILED);
			return response;
		} catch (XMLDBException e) {
			log.warning(e.getLocalizedMessage());
			response.addNewDoNotificationResponse().setStatus(
					MessageStatusType.MESSAGE_SENDING_FAILED);
			return response;
		}
		response.addNewDoNotificationResponse().setStatus(
				MessageStatusType.MESSAGE_SENDING_SUCCESSFULL);
		return response;
	}

	/**
	 * MAIN function to send a Communication
	 */
	public DoCommunicationResponseDocument doCommunication(
			DoCommunicationDocument dcd) throws WNSServiceException {
		DoCommunicationResponseDocument response = DoCommunicationResponseDocument.Factory
				.newInstance();
		org.x52North.wns.WNSUserDocument.WNSUser.SingleUser.User.Channel[] channel;
		XMLResource document;
		try {
			channel = userHandler.getMessageChannels(dcd.getDoCommunication()
					.getUserID());
			boolean storeMessage = false;
			for (int i = 0; i <= channel.length - 1; i++) {
				switch (channel[i].getProtocol().intValue()) {
				case org.x52North.wns.WNSUserDocument.WNSUser.SingleUser.User.Channel.Protocol.INT_EMAIL:
					if (mailHandler != null)
						mailHandler.sendCommunicationMessage(dcd, channel[i]
								.getTargetArray());
					break;
				case org.x52North.wns.WNSUserDocument.WNSUser.SingleUser.User.Channel.Protocol.INT_XMPP:
					if (xmppHandler != null)
						xmppHandler.sendCommunicationMessage(dcd, channel[i]
								.getTargetArray());
					break;
				case org.x52North.wns.WNSUserDocument.WNSUser.SingleUser.User.Channel.Protocol.INT_SMS:
					if (smsHandler != null)
						smsHandler.sendCommunicationMessage(dcd, channel[i]
								.getTargetArray(), "C" + messageID);
					storeMessage = true;
					break;
				case org.x52North.wns.WNSUserDocument.WNSUser.SingleUser.User.Channel.Protocol.INT_HTTP:
					if (wsHandler != null)
						wsHandler.sendCommunicationMessage(dcd, channel[i]
								.getTargetArray());
					break;
				case org.x52North.wns.WNSUserDocument.WNSUser.SingleUser.User.Channel.Protocol.INT_PHONE:
					if (phoneHandler != null)
						phoneHandler.sendCommunicationMessage(dcd, channel[i]
								.getTargetArray(), "C" + messageID);
					storeMessage = true;
					break;
				case org.x52North.wns.WNSUserDocument.WNSUser.SingleUser.User.Channel.Protocol.INT_FAX:
					if (faxHandler != null)
						faxHandler.sendCommunicationMessage(dcd, channel[i]
								.getTargetArray(), "C" + messageID);
					storeMessage = true;
					break;

				default:
					response.addNewDoCommunicationResponse().setStatus(
							MessageStatusType.MESSAGE_SENDING_FAILED);
					return response;

				}
			}
			if (storeMessage) {
				document = (XMLResource) col.createResource("N" + messageID,
						"XMLResource");
				document.setContent(dcd.getDoCommunication().getMessage()
						.xmlText());
				col.storeResource(document);
				col.close();
				if (dcd.getDoCommunication().getMaxTTLOfMessage()
						.compareToGDuration(maxTTLofMessage) < 0) {
					thread.add("N" + messageID, dcd.getDoCommunication()
							.getMaxTTLOfMessage());
				} else {
					thread.add("N" + messageID, maxTTLofMessage);
				}
				messageID++;
			}
		} catch (WNSException e) {
			response.getDoCommunicationResponse().setStatus(
					MessageStatusType.MESSAGE_SENDING_FAILED);
			return response;
		} catch (XMLDBException e) {
			log.warning(e.getLocalizedMessage());
		}
		response.addNewDoCommunicationResponse().setStatus(
				MessageStatusType.MESSAGE_SENDING_SUCCESSFULL);
		return response;

	}

	/**
	 * MAIN function to get the fully XML message
	 */
	public GetMessageResponseDocument getMessage(GetMessageDocument gmd)
			throws WNSServiceException {
		String mID = gmd.getGetMessage().getMessageID();
		GetMessageResponseDocument response = GetMessageResponseDocument.Factory
				.newInstance();

		XMLResource res;
		try {
			res = (XMLResource) col.getResource(mID);

			if (mID.substring(0, 1).equalsIgnoreCase("n")) {
				response.addNewGetMessageResponse().setNotificationMessage(
						NotificationMessageType.Factory.parse(res
								.getContentAsDOM()));
			} else {
				if (mID.substring(0, 1).equalsIgnoreCase("c")) {
					response.addNewGetMessageResponse()
							.setCommunicationMessage(
									CommunicationMessageType.Factory.parse(res
											.getContentAsDOM()));
				} else {
					WNSServiceException s = new WNSServiceException(
							WNSServiceException.ExceptionLevel.PlainExceptions);
					s.addCodedException(
							WNSServiceException.ExceptionCode.ExpiredMessageID,
							null);
					throw s;
				}
			}
		} catch (Exception e) {
			WNSServiceException s = new WNSServiceException(
					WNSServiceException.ExceptionLevel.PlainExceptions);
			s.addCodedException(
					WNSServiceException.ExceptionCode.ExpiredMessageID, null);
			throw s;
		}
		return response;
	}

}
