/******************************************************************************
 Web Notification Service (WNS): An open interface for a service
 by which a client may conduct asynchronous dialogues
 (message interchanges) with one or more other services.
 *******************************************************************************/

package org.n52.wns;

import java.io.IOException;
import java.io.InputStream;
import java.io.StringWriter;
import java.util.ArrayList;
import java.util.Enumeration;
import java.util.Iterator;
import java.util.logging.Logger;

import javax.servlet.http.HttpServletRequest;
import javax.servlet.http.HttpServletResponse;

import net.opengeospatial.ows.CapabilitiesBaseType;
import net.opengeospatial.ows.SectionsType;
import net.opengis.wns.CapabilitiesDocument;
import net.opengis.wns.DoCommunicationDocument;
import net.opengis.wns.DoCommunicationResponseDocument;
import net.opengis.wns.DoNotificationDocument;
import net.opengis.wns.DoNotificationResponseDocument;
import net.opengis.wns.GetCapabilitiesDocument;
import net.opengis.wns.GetMessageDocument;
import net.opengis.wns.GetMessageResponseDocument;
import net.opengis.wns.ProtocolType;
import net.opengis.wns.ProtocolsType;
import net.opengis.wns.RegisterDocument;
import net.opengis.wns.RegisterUserResponseDocument;
import net.opengis.wns.UnregisterDocument;
import net.opengis.wns.UnregisterResponseDocument;
import net.opengis.wns.UpdateMultiUserRegistrationDocument;
import net.opengis.wns.UpdateMultiUserRegistrationResponseDocument;
import net.opengis.wns.UpdateSingleUserRegistrationDocument;
import net.opengis.wns.UpdateSingleUserRegistrationResponseDocument;
import net.opengis.wns.CapabilitiesDocument.Capabilities;
import net.opengis.wns.ContentsDocument.Contents;
import net.opengis.wns.GetCapabilitiesDocument.GetCapabilities;

import org.apache.xmlbeans.SchemaType;
import org.apache.xmlbeans.XmlException;
import org.apache.xmlbeans.XmlObject;
import org.apache.xmlbeans.XmlOptions;
import org.n52.wns.WNSServiceException.ExceptionLevel;
import org.n52.wns.db.DAOFactory;
import org.n52.wns.db.UserDAO;
import org.x52North.wns.WNSUserDocument;
import org.x52North.wns.WNSConfigDocument.WNSConfig;

/**
 * RequestHandler handles all incoming requests
 * 
 * @author Johannes Echterhoff, Dennis Dahlmann
 * @version 1.0
 */

public class WNSRequestHandler {

	private static Logger logger = Logger.getLogger(WNSRequestHandler.class
			.getName());

	private String ogcExcSchemaLocation;

	private String wnsSchemaLocation;

	private ExceptionLevel excLevel;

	private WNSConfig config;

	public WNSRequestHandler() {
		WNSInitParamContainer initParams = WNSInitParamContainer.getInstance();

		config = initParams.getWnsConfig();

		ogcExcSchemaLocation = initParams.getOgcExceptionSchemaLocation();
		wnsSchemaLocation = initParams.getWnsSchemaLocation();
		excLevel = initParams.getExceptionLevel();
	}

	public void handleGetRequest(HttpServletRequest request,
			HttpServletResponse response) throws WNSException {

		// log the request
		logger.fine("Connected from: " + request.getRemoteAddr() + " "
				+ request.getRemoteHost() + "\tRequest: ["
				+ request.getQueryString() + "]\n");

		WNSServiceException se = new WNSServiceException(excLevel);
		boolean requestFound = false;

		Enumeration params = request.getParameterNames();
		while (params.hasMoreElements()) {

			// search for the request-parameter to determine which operation
			// shall be performed
			String p = (String) params.nextElement();

			if (p.equalsIgnoreCase(WNSConstants.REQUEST)) {
				requestFound = true;
				String paramReq = request.getParameter(p);

				if (paramReq.equalsIgnoreCase(WNSConstants.LISTWNSUSER)) {

					WNSInitParamContainer initParams = WNSInitParamContainer
							.getInstance();
					DAOFactory daofac = initParams.getDAOFactory();
					UserDAO userdao = daofac.getUserDAO();
					WNSUserDocument wnsud = userdao.getUserDocument();
					WNSUtil.sendResponse(wnsud, null, response);

				} else if (paramReq
						.equalsIgnoreCase(WNSConstants.GETCAPABILITIES)) {

					// check parameter SERVICE
					String service = WNSUtil.getValue4Param("Service", request);
					if (service == null) {
						se
								.addCodedException(
										WNSServiceException.ExceptionCode.MissingParameterValue,
										"SERVICE",
										"The mandatory parameter 'SERVICE' was not fo"
												+ "und in the request");
						WNSUtil.sendResponse(se.getDocument(),
								ogcExcSchemaLocation, response);

					} else if (!service.equalsIgnoreCase(WNSConstants.SERVICE)) {
						se
								.addCodedException(
										WNSServiceException.ExceptionCode.InvalidParameterValue,
										"SERVICE",
										"The value of the mandatory parameter 'SERVICE'"
												+ "must be '"
												+ WNSConstants.SERVICE
												+ "'. Delivered value was: "
												+ service);
						WNSUtil.sendResponse(se.getDocument(),
								ogcExcSchemaLocation, response);
					}

					// TODO check version if necessary - for now version is
					// ignored

					// check parameter SECTIONS
					String sections = WNSUtil.getValue4Param("Sections",
							request);
					if (sections != null) {

						boolean all = false;
						boolean serviceIdentification = false;
						boolean serviceProvider = false;
						boolean operationsMetadata = false;
						boolean contents = false;

						boolean nothing = true;

						String[] tokens = sections.split(",");
						for (int i = 0; i < tokens.length; i++) {

							if (tokens[i].trim().equalsIgnoreCase("all")) {
								nothing = false;
								all = true;
								break;
							} else if (tokens[i].trim().equalsIgnoreCase(
									"serviceIdentification")) {
								nothing = false;
								serviceIdentification = true;
							} else if (tokens[i].trim().equalsIgnoreCase(
									"serviceProvider")) {
								nothing = false;
								serviceProvider = true;
							} else if (tokens[i].trim().equalsIgnoreCase(
									"operationsMetadata")) {
								nothing = false;
								operationsMetadata = true;
							} else if (tokens[i].trim().equalsIgnoreCase(
									"contents")) {
								nothing = false;
								contents = true;
							}
						}

						if (nothing || all) {
							// simply return all
							GetCapabilitiesDocument gcd = GetCapabilitiesDocument.Factory
									.newInstance();
							GetCapabilities gc = gcd.addNewGetCapabilities();
							gc.setService(WNSConstants.SERVICE);

							try {
								this.handleGetCapabilities(gcd, response);
							} catch (Exception e) {
								throw new WNSException(
										"Could not handle GetCapabilities request.",
										e);
							}
						} else {

							// create custom CapabilitiesResponse
							GetCapabilitiesDocument gcd = GetCapabilitiesDocument.Factory
									.newInstance();
							GetCapabilities gc = gcd.addNewGetCapabilities();
							gc.setService(WNSConstants.SERVICE);
							SectionsType st = gc.addNewSections();

							if (serviceIdentification)
								st.addSection("serviceIdentification");
							if (serviceProvider)
								st.addSection("serviceProvider");
							if (operationsMetadata)
								st.addSection("operationsMetadata");
							if (contents)
								st.addSection("contents");

							try {
								this.handleGetCapabilities(gcd, response);
							} catch (Exception e) {
								throw new WNSException(
										"Could not handle GetCapabilities request.",
										e);
							}
						}

					} else {
						// return whole document
						GetCapabilitiesDocument gcd = GetCapabilitiesDocument.Factory
								.newInstance();
						GetCapabilities gc = gcd.addNewGetCapabilities();
						gc.setService(WNSConstants.SERVICE);

						try {
							this.handleGetCapabilities(gcd, response);
						} catch (Exception e) {
							throw new WNSException(
									"Could not handle GetCapabilities request.",
									e);
						}
					}
				} else {
					se
							.addCodedException(
									WNSServiceException.ExceptionCode.OperationNotSupported,
									null,
									"The requested operation '"
											+ paramReq
											+ "' is"
											+ " not supported by the HTTP-GET binding");
					WNSUtil.sendResponse(se.getDocument(),
							ogcExcSchemaLocation, response);
				}
			}
		}

		if (!requestFound) {
			se.addCodedException(
					WNSServiceException.ExceptionCode.MissingParameterValue,
					"REQUEST",
					"The mandatory parameter 'REQUEST' was not found in the "
							+ "request");
			WNSUtil.sendResponse(se.getDocument(), ogcExcSchemaLocation,
					response);
		}
	}

	public void handleGetCapabilities(GetCapabilitiesDocument gcd,
			HttpServletResponse response) throws WNSException {

		GetCapabilities gc = gcd.getGetCapabilities();
		WNSServiceException se = new WNSServiceException(excLevel);

		if (!gc.getService().equals(WNSConstants.SERVICE)) {
			se.addCodedException(
					WNSServiceException.ExceptionCode.InvalidParameterValue,
					"SERVICE", "You did not request the service '"
							+ WNSConstants.SERVICE + "'.");
			WNSUtil.sendResponse(se.getDocument(), ogcExcSchemaLocation,
					response);

		} else {

			if (gc.isSetSections()) {

				boolean all = false;
				boolean serviceIdentification = false;
				boolean serviceProvider = false;
				boolean operationsMetadata = false;
				boolean contents = false;
				boolean nothing = true;

				SectionsType st = gc.getSections();
				String[] sections = st.getSectionArray();
				for (int i = 0; i < sections.length; i++) {
					if (sections[i].trim().equalsIgnoreCase("all")) {
						nothing = false;
						all = true;
						break;
					} else if (sections[i].trim().equalsIgnoreCase(
							"serviceIdentification")) {
						nothing = false;
						serviceIdentification = true;
					} else if (sections[i].trim().equalsIgnoreCase(
							"serviceProvider")) {
						nothing = false;
						serviceProvider = true;
					} else if (sections[i].trim().equalsIgnoreCase(
							"operationsMetadata")) {
						nothing = false;
						operationsMetadata = true;
					} else if (sections[i].trim().equalsIgnoreCase("contents")) {
						nothing = false;
						contents = true;
					}
				}

				if (nothing || all) {
					// simply return all
					sendCompleteCapabilities(response);
				} else {

					CapabilitiesBaseType cbt = config.getServiceProperties()
							.getCapabilitiesBaseInformation();

					// create custom CapabilitiesResponse
					CapabilitiesDocument capsd = CapabilitiesDocument.Factory
							.newInstance();
					Capabilities caps = capsd.addNewCapabilities();

					caps.setVersion(cbt.getVersion());
					caps.setUpdateSequence(cbt.getUpdateSequence());
					if (serviceIdentification)
						caps.setServiceIdentification(cbt
								.getServiceIdentification());
					if (serviceProvider)
						caps.setServiceProvider(cbt.getServiceProvider());
					if (operationsMetadata)
						caps.setOperationsMetadata(cbt.getOperationsMetadata());
					if (contents)
						caps.setContents(this.createContents());

					WNSUtil.sendResponse(capsd, wnsSchemaLocation, response);
				}

			} else {
				// return whole document
				sendCompleteCapabilities(response);
			}
		}
	}

	private void sendCompleteCapabilities(HttpServletResponse response)
			throws WNSException {

		CapabilitiesBaseType cbt = config.getServiceProperties()
				.getCapabilitiesBaseInformation();

		CapabilitiesDocument capsd = CapabilitiesDocument.Factory.newInstance();
		Capabilities caps = capsd.addNewCapabilities();

		caps.setVersion(cbt.getVersion());
		caps.setUpdateSequence(cbt.getUpdateSequence());
		caps.setServiceIdentification(cbt.getServiceIdentification());
		caps.setServiceProvider(cbt.getServiceProvider());
		caps.setOperationsMetadata(cbt.getOperationsMetadata());
		caps.setContents(this.createContents());

		WNSUtil.sendResponse(capsd, wnsSchemaLocation, response);
	}

	private Contents createContents() {

		Contents c = Contents.Factory.newInstance();
		ProtocolsType pt = c.addNewSupportedCommunicationProtocols();

		if (config.getRegisteredHandlers().isSetXMPPHandler()) {
			pt.addProtocol(ProtocolType.XMPP);
		}
		if (config.getRegisteredHandlers().isSetFaxHandler()) {
			pt.addProtocol(ProtocolType.FAX);
		}
		if (config.getRegisteredHandlers().isSetEMailHandler()) {
			pt.addProtocol(ProtocolType.EMAIL);
		}
		if (config.getRegisteredHandlers().isSetPhoneHandler()) {
			pt.addProtocol(ProtocolType.PHONE);
		}
		if (config.getRegisteredHandlers().isSetSMSHandler()) {
			pt.addProtocol(ProtocolType.SMS);
		}

		c.setMaxTTLOfMessages(config.getServiceProperties()
				.getMaxTTLOfMessages());

		return c;
	}

	public void handlePostRequest(HttpServletRequest request,
			HttpServletResponse response) throws WNSException {

		logger.info("Connected from: " + request.getRemoteAddr() + " "
				+ request.getRemoteHost());

		try {

			String input = null;

			try {
				// Read the request
				InputStream in = request.getInputStream();
				String inputString = "";
				String decodedString = "";
				// read every byte of the request and put it into a StringWriter
				byte[] buffer = new byte[8192];
				StringWriter sw = new StringWriter();
				int bytesRead;
				while ((bytesRead = in.read(buffer)) != -1) {
					sw.write(new String(buffer, 0, bytesRead));
				}
				inputString = sw.toString();
				// discard "request="-Input String header
				if (inputString.startsWith("request=")) {
					// discard "request="-Input String header
					inputString = inputString
							.substring(8, inputString.length());

					// decode the application/x-www-form-urlencoded query string
					decodedString = java.net.URLDecoder.decode(inputString,
							"UTF-8");
					input = decodedString;
				} else {
					input = inputString;
				}

				logger.info("Request is: " + input);

			} catch (IOException e) {
				WNSServiceException se = new WNSServiceException(excLevel);
				se.addCodedException(
						WNSServiceException.ExceptionCode.NoApplicableCode,
						null, e);
				throw se;
			}

			// parse request and handle accordingly
			XmlObject xobj = null;
			ArrayList parsingErrors = new ArrayList();
			XmlOptions parsingOptions = new XmlOptions();
			parsingOptions.setErrorListener(parsingErrors);

			try {

				xobj = XmlObject.Factory.parse(input, parsingOptions);

			} catch (XmlException e) {
				throw this.createInvalidRequestException(parsingErrors,
						excLevel, ogcExcSchemaLocation, response);
			}

			// handle request
			ArrayList validationErrors = new ArrayList();
			XmlOptions validationOptions = new XmlOptions();
			validationOptions.setErrorListener(validationErrors);
			boolean isValid = true;

			WNSUserHandler userhandler = WNSUserHandler.getInstance();

			WNSMessageHandler messageHandler = WNSMessageHandler
					.getInstance(userhandler);

			SchemaType type = xobj.schemaType();

			if (type == GetCapabilitiesDocument.type) {

				GetCapabilitiesDocument gcd = (GetCapabilitiesDocument) xobj;
				isValid = gcd.validate(validationOptions);
				if (!isValid) {
					throw this.createInvalidRequestException(validationErrors,
							excLevel, ogcExcSchemaLocation, response);
				} else {
					this.handleGetCapabilities(gcd, response);
				}

			} else if (type == RegisterDocument.type) {

				RegisterDocument rd = (RegisterDocument) xobj;
				isValid = rd.validate(validationOptions);
				if (!isValid) {
					throw this.createInvalidRequestException(validationErrors,
							excLevel, ogcExcSchemaLocation, response);
				} else {
					RegisterUserResponseDocument rurd = userhandler
							.register(rd);
					WNSUtil.sendResponse(rurd, wnsSchemaLocation, response);
				}

			} else if (type == UpdateSingleUserRegistrationDocument.type) {

				UpdateSingleUserRegistrationDocument usurd = (UpdateSingleUserRegistrationDocument) xobj;
				isValid = usurd.validate(validationOptions);
				if (!isValid) {
					throw this.createInvalidRequestException(validationErrors,
							excLevel, ogcExcSchemaLocation, response);
				} else {
					UpdateSingleUserRegistrationResponseDocument usurrd = userhandler
							.updateSingleUser(usurd);
					WNSUtil.sendResponse(usurrd, wnsSchemaLocation, response);
				}

			} else if (type == UpdateMultiUserRegistrationDocument.type) {

				UpdateMultiUserRegistrationDocument umurd = (UpdateMultiUserRegistrationDocument) xobj;
				isValid = umurd.validate(validationOptions);
				if (!isValid) {
					throw this.createInvalidRequestException(validationErrors,
							excLevel, ogcExcSchemaLocation, response);
				} else {
					UpdateMultiUserRegistrationResponseDocument umurrd = userhandler
							.updateMultiUser(umurd);
					WNSUtil.sendResponse(umurrd, wnsSchemaLocation, response);
				}

			} else if (type == UnregisterDocument.type) {

				UnregisterDocument ud = (UnregisterDocument) xobj;
				isValid = ud.validate(validationOptions);
				if (!isValid) {
					throw this.createInvalidRequestException(validationErrors,
							excLevel, ogcExcSchemaLocation, response);
				} else {
					UnregisterResponseDocument urd = userhandler.unregister(ud);
					WNSUtil.sendResponse(urd, wnsSchemaLocation, response);
				}

			} else if (type == DoNotificationDocument.type) {

				DoNotificationDocument dnd = (DoNotificationDocument) xobj;
				isValid = dnd.validate(validationOptions);
				if (!isValid) {
					throw this.createInvalidRequestException(validationErrors,
							excLevel, ogcExcSchemaLocation, response);
				} else {
					DoNotificationResponseDocument dnrd = messageHandler
							.doNotification(dnd);
					WNSUtil.sendResponse(dnrd, wnsSchemaLocation, response);
				}

			} else if (type == DoCommunicationDocument.type) {

				DoCommunicationDocument dcd = (DoCommunicationDocument) xobj;
				isValid = dcd.validate(validationOptions);
				if (!isValid) {
					throw this.createInvalidRequestException(validationErrors,
							excLevel, ogcExcSchemaLocation, response);
				} else {
					DoCommunicationResponseDocument dcrd = messageHandler
							.doCommunication(dcd);
					WNSUtil.sendResponse(dcrd, wnsSchemaLocation, response);
				}

			} else if (type == GetMessageDocument.type) {

				GetMessageDocument gmd = (GetMessageDocument) xobj;
				isValid = gmd.validate(validationOptions);
				if (!isValid) {
					throw this.createInvalidRequestException(validationErrors,
							excLevel, ogcExcSchemaLocation, response);
				} else {
					GetMessageResponseDocument gmrd = messageHandler
							.getMessage(gmd);
					WNSUtil.sendResponse(gmrd, wnsSchemaLocation, response);
				}

			} else {
				// unknown request
				WNSServiceException se = new WNSServiceException(excLevel);
				se
						.addCodedException(
								WNSServiceException.ExceptionCode.OperationNotSupported,
								null, "The operation you requested ("
										+ type.getName().toString()
										+ ") is not supported by "
										+ "this service");
				throw se;
			}

		} catch (WNSServiceException se) {
			WNSUtil.sendResponse(se.getDocument(), ogcExcSchemaLocation,
					response);
		}
	}

	private WNSServiceException createInvalidRequestException(ArrayList errors,
			ExceptionLevel excLevel, String owsSchemaLocation,
			HttpServletResponse response) {
		StringBuffer sb = new StringBuffer();
		Iterator iter = errors.iterator();
		while (iter.hasNext()) {
			sb.append("[Error] " + iter.next() + "\n");
		}

		WNSServiceException se = new WNSServiceException(excLevel);
		se
				.addCodedException(
						WNSServiceException.ExceptionCode.InvalidRequest, sb
								.toString(),
						"Your request was invalid. Examine locator-element for more details.");

		return se;
	}
}
