/******************************************************************************
 Web Notification Service (WNS): An open interface for a service
 by which a client may conduct asynchronous dialogues
 (message interchanges) with one or more other services.
 *******************************************************************************/
package org.n52.wns;

import java.util.ArrayList;
import java.util.logging.Logger;

import net.opengeospatial.ows.ExceptionDocument;
import net.opengeospatial.ows.ExceptionReportDocument;
import net.opengeospatial.ows.ExceptionType;
import net.opengeospatial.ows.ExceptionReportDocument.ExceptionReport;

/**
 * Exception class
 * 
 * @author Johannes Echterhoff, Dennis Dahlmann
 * @version 1.0
 * 
 */
public class WNSServiceException extends Exception {

	public enum ExceptionCode {
		// standard OGC codes
		InvalidParameterValue, MissingParameterValue, NoApplicableCode, OperationNotSupported,

		// WNS codes
		UnknownUserID, InvalidRequest, InvalidUserID, ExpiredMessageID, DatabaseErrorOccured, XMPPServerLoginError, EmailHandlerError, ErrorDuringStoringUserFile;
	}

	public enum ExceptionLevel {
		DetailedExceptions, PlainExceptions
	}

	private static Logger logger = Logger.getLogger(WNSServiceException.class
			.getName());

	private ExceptionLevel excLevel = null;

	private ArrayList<ExceptionType> excs = new ArrayList<ExceptionType>();

	public WNSServiceException(ExceptionLevel excLevelIn) {
		this.excLevel = excLevelIn;
	}

	public void addCodedException(ExceptionCode code, String locator) {

		ExceptionType et = ExceptionType.Factory.newInstance();
		et.setExceptionCode(code.toString());

		if (locator != null)
			et.setLocator(locator);

		excs.add(et);
	}

	public void addCodedException(ExceptionCode code, String locator,
			Exception e) {

		ExceptionType et = ExceptionType.Factory.newInstance();
		et.setExceptionCode(code.toString());
		if (locator != null)
			et.setLocator(locator);

		String name = e.getClass().getName();
		String message = e.getMessage();
		StackTraceElement[] stackTraces = e.getStackTrace();

		StringBuffer sb = new StringBuffer();
		sb.append("[EXC] internal service exception");
		if (excLevel.compareTo(ExceptionLevel.PlainExceptions) == 0) {
			sb.append(". Message: " + message);
		} else if (excLevel.compareTo(ExceptionLevel.DetailedExceptions) == 0) {
			sb.append(": " + name + "\n");
			sb.append("[EXC] message: " + message + "\n");
			for (int i = 0; i < stackTraces.length; i++) {
				StackTraceElement element = stackTraces[i];
				sb.append("[EXC] " + element.toString() + "\n");
			}
		} else {
			logger.warning("addCodedException: unknown ExceptionLevel " + "("
					+ excLevel.toString() + ")occurred.");
		}

		et.addExceptionText(sb.toString());
		// TODO i guess there is a better way to format an exception

		excs.add(et);
	}

	public void addCodedException(ExceptionCode code, String locator,
			String message) {

		ExceptionType et = ExceptionType.Factory.newInstance();
		et.setExceptionCode(code.toString());

		if (locator != null)
			et.setLocator(locator);
		if (message != null)
			et.addExceptionText(message);

		excs.add(et);
	}

	public void addCodedException(ExceptionCode code, String locator,
			String[] messages) {
		ExceptionType et = ExceptionType.Factory.newInstance();
		et.setExceptionCode(code.toString());
		if (locator != null)
			et.setLocator(locator);
		for (int i = 0; i < messages.length; i++) {
			String string = messages[i];
			et.addExceptionText(string);
		}
		excs.add(et);
	}

	public void addExceptionReport(ExceptionDocument ed) {
		this.excs.add(ed.getException());
	}

	public void addExceptionReport(ExceptionReportDocument erd) {
		for (ExceptionType et : erd.getExceptionReport().getExceptionArray()) {
			this.excs.add(et);
		}
	}

	public void addServiceException(WNSServiceException seIn) {
		this.excs.addAll(seIn.getExceptions());
	}

	public boolean containsCode(ExceptionCode ec) {
		for (ExceptionType et : excs) {
			if (et.getExceptionCode().equalsIgnoreCase(ec.toString()))
				return true;
		}
		return false;
	}

	public boolean containsExceptions() {
		return excs.size() > 0;
	}

	public ExceptionReportDocument getDocument() {

		ExceptionReportDocument erd = ExceptionReportDocument.Factory
				.newInstance();
		ExceptionReport er = ExceptionReport.Factory.newInstance();
		er.setLanguage("en");
		er.setVersion(WNSConstants.OWSEXCEPTIONREPORTVERSION); // TODO what
		// version??
		// insert the
		// number
		// dynamically??
		er.setExceptionArray(excs.toArray(new ExceptionType[excs.size()]));
		erd.setExceptionReport(er);

		return erd;
	}

	public ArrayList<ExceptionType> getExceptions() {
		return excs;
	}

	public String getMessage() {
		return this.getDocument().toString();
	}
}
