/******************************************************************************
 Web Notification Service (WNS): An open interface for a service
 by which a client may conduct asynchronous dialogues
 (message interchanges) with one or more other services.
 *******************************************************************************/
package org.n52.wns;

import java.io.IOException;
import java.util.Enumeration;
import java.util.logging.Level;
import java.util.logging.Logger;

import javax.servlet.ServletOutputStream;
import javax.servlet.http.HttpServletRequest;
import javax.servlet.http.HttpServletResponse;
import javax.xml.namespace.QName;

import org.apache.xmlbeans.XmlCursor;
import org.apache.xmlbeans.XmlOptions;
import org.apache.xmlbeans.XmlTokenSource;

/**
 * @author Johannes Echterhoff
 * @version 1.0
 * 
 */
public class WNSUtil {

	private static Logger logger = Logger.getLogger(WNSUtil.class.getName());

	/**
	 * Returns the parameter value for the passed parameter. This method is used
	 * to provide caseinsensitive parameter spelling.
	 * 
	 * @param param
	 *            the parameter, for which the parameter value schould be
	 *            returned
	 * @param req
	 *            the request from which the parameter value should be extracted
	 * @return returns the parameter value for the passed parameter (the search
	 *         for the parameter is caseinsensitive) or null if the given
	 *         parameter does not exist
	 */
	public static String getValue4Param(String param, HttpServletRequest req) {
		String paramValue = null;
		Enumeration params = req.getParameterNames();
		while (params.hasMoreElements()) {
			String p = (String) params.nextElement();
			if (p.equalsIgnoreCase(param)) {
				return req.getParameter(p);
			}
		}
		return paramValue;
	}

	/**
	 * Sends the given XML to via the given response object. The given schema
	 * location is added to the XML.
	 * 
	 * The XML will be pretty printed.
	 * 
	 * @param xts
	 *            Contains the XML encoded response.
	 * @param schemaLocation
	 *            Will be added to the XML if not null.
	 * @param response
	 *            Will deliver the XML.
	 * @throws SPSException
	 *             If the response could not be sent.
	 */
	public static void sendResponse(XmlTokenSource xts, String schemaLocation,
			HttpServletResponse response) throws WNSException {

		response.setContentType("text/xml");
		ServletOutputStream out = null;
		try {
			out = response.getOutputStream();

			XmlOptions options = new XmlOptions();

			options.setSavePrettyPrint();
			options.setUseDefaultNamespace();
			options.setSaveAggressiveNamespaces();

			// If the schemaLocation is null do not insert schemaLocation to
			// xml-instance
			if (schemaLocation != null && schemaLocation.trim().length() > 0) {
				XmlCursor c = xts.newCursor();
				c.toFirstChild();
				c.setAttributeText(new QName(
						"http://www.w3.org/2001/XMLSchema-instance",
						"schemaLocation", "xsi"), schemaLocation);
				c.dispose();
			}

			xts.save(out, options);

		} catch (IOException e) {
			throw new WNSException("Could not send response ", e);
		} finally {
			if (out != null) {
				try {
					out.flush();
					out.close();
				} catch (IOException e) {
					logger.log(Level.WARNING,
							"Cannot close response output stream.", e);
				}
			}
		}
	}

}
