/******************************************************************************
 Web Notification Service (WNS): An open interface for a service
 by which a client may conduct asynchronous dialogues
 (message interchanges) with one or more other services.
 *******************************************************************************/
package org.n52.wns.xmpp;

import java.util.logging.Logger;

import net.opengis.wns.CommunicationMessageDocument;
import net.opengis.wns.DoCommunicationDocument;
import net.opengis.wns.DoNotificationDocument;
import net.opengis.wns.NotificationMessageDocument;

import org.jivesoftware.smack.SASLAuthentication;
import org.jivesoftware.smack.SSLXMPPConnection;
import org.jivesoftware.smack.XMPPConnection;
import org.jivesoftware.smack.XMPPException;
import org.n52.wns.WNSException;
import org.n52.wns.WNSServiceException;
import org.n52.wns.WNSServiceException.ExceptionCode;
import org.n52.wns.WNSServiceException.ExceptionLevel;
import org.x52North.wns.WNSConfigDocument;
import org.x52North.wns.WNSConfigDocument.WNSConfig.RegisteredHandlers.XMPPHandler;

/**
 * XMPPHandlerJabJab organizes the sending of instant messages with XMPP via the
 * jabjab.de server
 * 
 * @author Dennis Dahlmann
 * @version 1.0
 */
public class XMPPHandlerJabber implements org.n52.wns.xmpp.XMPPHandler {

	private Logger log = Logger.getLogger(XMPPHandlerJabber.class.getName());

	private XMPPConnection con;

	private String user;

	private String pwd;

	private String server = "";

	private int port = 5222; // default Jabber port

	private static XMPPHandlerJabber instance = null;

	private XMPPHandlerJabber(XMPPHandler xmppHandler)
			throws WNSServiceException {

		this.setProperties(xmppHandler);

	}

	public static XMPPHandlerJabber getInstance(
			WNSConfigDocument.WNSConfig.RegisteredHandlers.XMPPHandler xmppHandler)
			throws WNSServiceException {
		if (instance == null) {
			return instance = new XMPPHandlerJabber(xmppHandler);
		} else {
			return instance;
		}

	}

	// /**
	// * protected constructor sets up the XMPP environment only used for
	// * debugging!
	// */
	// protected XMPPHandlerJabber(String user, String pwd, String server)
	// throws WNSException {
	// this.user = user;
	// this.pwd = pwd;
	// this.server = server;
	// try {
	// // set up the xmpp environment
	// this.setup();
	// } catch (Exception e) {
	// log
	// .severe("Error occured during setting up the XMPP environment. XMPP
	// disabled!");
	// e.printStackTrace();
	// if (e instanceof WNSException)
	// throw new WNSException(e.getMessage());
	// else
	// throw new WNSException(
	// "Error occured during setting up the XMPP environment");
	// }
	// }

	/**
	 * Setting up the properties
	 */
	public void setProperties(
			WNSConfigDocument.WNSConfig.RegisteredHandlers.XMPPHandler xmppHandler)
			throws WNSServiceException {

		this.user = xmppHandler.getUser();
		this.pwd = xmppHandler.getPasswd();
		this.port = xmppHandler.getPort();
		this.server = xmppHandler.getProvider();

		try {
			this.setup();
			this.login();
		} catch (Exception ex) {
			WNSServiceException service = new WNSServiceException(
					ExceptionLevel.PlainExceptions);
			service.addCodedException(ExceptionCode.XMPPServerLoginError,
					"XMPPHandlerJabber, setProperties()");
			throw service;
		}
	}

	/**
	 * set up the Instant Messsage environment, called from setProperties
	 */
	private void setup() throws WNSException {
		try {
			con = new XMPPConnection(server, port);
		} catch (XMPPException e) {
			try {
				con = new SSLXMPPConnection(server, port);
			} catch (Exception e1) {
				throw new WNSException("Failure to login to XMPP server");
			}

		}
	}

	/**
	 * login creates a connection to the xmpp server and tries to login the
	 * user, called from setProperties
	 */
	private void login() throws WNSException {
		try {
			SASLAuthentication auth = con.getSASLAuthentication();
			auth.authenticate(user, pwd, "");
		} catch (XMPPException e) {
			log.warning(e.getMessage());
			throw new WNSException("Login to XMPP server failed ");
		}
	}

	/**
	 * Returns the status of the connection
	 */
	public boolean isConnected() {
		return con.isConnected();
	}

	/**
	 * sends a private message to the server.
	 * 
	 * @param String
	 *            recipient of the message
	 * @param String
	 *            message to be sent
	 */
	public void sendMessage(String recipient, String message)
			throws WNSException {
		try {
			con.createChat(recipient).sendMessage(message);
		} catch (XMPPException e1) {
			log.warning(e1.getMessage());
		}
	}

	/**
	 * sends a Notification Message. Does not wait until the message is replied
	 * 
	 * @param WNSUser
	 *            user to send the IM to
	 * @param WNSMessage
	 *            message to be sent
	 */
	public void sendNotificationMessage(DoNotificationDocument dnd,
			String[] targets) throws WNSServiceException {
		String bodyText;
		// get the address
		for (int i = 0; i <= targets.length - 1; i++) {
			String recipient = targets[i];
			NotificationMessageDocument notdoc = NotificationMessageDocument.Factory
					.newInstance();
			notdoc.setNotificationMessage(dnd.getDoNotification().getMessage());
			bodyText = notdoc.toString();
			try {
				sendMessage(recipient, (bodyText.toString()));
			} catch (WNSException e) {
				log.warning(e.getMessage());
			}
		}

	}

	/**
	 * sends a Communication Message, does not wait until the message ist
	 * replied
	 * 
	 * @param WNSUser
	 *            user to send the IM to
	 * @param WNSMessage
	 *            message to be sent
	 */
	public void sendCommunicationMessage(DoCommunicationDocument dcd,
			String[] targets) throws WNSServiceException {
		String bodyText;
		// get the address
		for (int i = 0; i <= targets.length - 1; i++) {
			String recipient = targets[i].toString();
			CommunicationMessageDocument comdoc = CommunicationMessageDocument.Factory
					.newInstance();
			comdoc.setCommunicationMessage(dcd.getDoCommunication()
					.getMessage());
			bodyText = comdoc.toString();
			try {
				sendMessage(recipient, (bodyText.toString()));
			} catch (WNSException e) {
				log.warning(e.getMessage());
			}
		}
	}
}
